"""Contains the `SlepianIdentity` class."""

import numpy as np
import numpy.typing as npt
import pydantic
import typing_extensions

import sleplet._string_methods
import sleplet._validation
from sleplet.functions.fp import Fp


@pydantic.dataclasses.dataclass(config=sleplet._validation.validation)
class SlepianIdentity(Fp):
    """Create an identify function in the Slepian region."""

    def __post_init__(self: typing_extensions.Self) -> None:
        super().__post_init__()

    def _create_coefficients(
        self: typing_extensions.Self,
    ) -> npt.NDArray[np.complex128 | np.float64]:
        return np.ones(self.L**2, dtype=np.complex128)

    def _create_name(self: typing_extensions.Self) -> str:
        return (
            f"{sleplet._string_methods._convert_camel_case_to_snake_case(self.__class__.__name__)}"
            f"_{self.slepian.region._name_ending}"
        )

    def _set_reality(self: typing_extensions.Self) -> bool:
        return False

    def _set_spin(self: typing_extensions.Self) -> int:
        return 0

    def _setup_args(self: typing_extensions.Self) -> None:
        if isinstance(self.extra_args, list):
            msg = f"{self.__class__.__name__} does not support extra arguments"
            raise TypeError(msg)



"""Methods to help in creating plots."""

import logging

import matplotlib as mpl
import numpy as np
import numpy.typing as npt

import pyssht as ssht

import sleplet._mask_methods
import sleplet._vars
import sleplet.functions.coefficients
import sleplet.harmonic_methods
import sleplet.meshes.mesh_coefficients
import sleplet.slepian.region
import sleplet.slepian_methods

_logger = logging.getLogger(__name__)


def calc_plot_resolution(L: int) -> int:
    """
    Calculate appropriate resolution for given L.

    Args:
        L: The spherical harmonic bandlimit.

    Returns:
        The output spherical harmonic bandlimit to boost the resolution.
    """
    res_dict = {1: 6, 2: 5, 3: 4, 7: 3, 9: 2, 10: 1}

    return next(
        (
            L * 2**exponent
            for log_bandlimit, exponent in res_dict.items()
            if 2**log_bandlimit > L
        ),
        L,
    )


def _convert_colourscale(
    cmap: mpl.colors.LinearSegmentedColormap,
    *,
    pl_entries: int = 255,
) -> list[tuple[float, str]]:
    """Convert cmocean colourscale to a plotly colourscale."""
    h = 1 / (pl_entries - 1)
    pl_colorscale = []

    for k in range(pl_entries):
        C = list(map(np.uint8, np.array(cmap(k * h)[:3]) * 255))
        pl_colorscale.append((k * h, f"rgb{(C[0], C[1], C[2])}"))

    return pl_colorscale


def _calc_nearest_grid_point(
    L: int,
    alpha_pi_fraction: float,
    beta_pi_fraction: float,
) -> tuple[float, float]:
    """
    Calculate nearest index of alpha/beta for translation
    this is due to calculating omega' through the pixel
    values - the translation needs to be at the same position
    as the rotation such that the difference error is small.
    """
    thetas, phis = ssht.sample_positions(L, Method=sleplet._vars.SAMPLING_SCHEME)
    pix_j = np.abs(phis - alpha_pi_fraction * np.pi).argmin()
    pix_i = np.abs(thetas - beta_pi_fraction * np.pi).argmin()
    alpha, beta = phis[pix_j], thetas[pix_i]
    msg = f"grid point: (alpha, beta)=({alpha:e}, {beta:e})"
    _logger.info(msg)
    return alpha, beta


def find_max_amplitude(
    function: sleplet.functions.coefficients.Coefficients,
    *,
    plot_type: str = "real",
    upsample: bool = True,
) -> float:
    """
    For a given set of coefficients it finds the largest absolute value for a
    given plot type such that plots can have the same scale as the input.

    Args:
        function: The `Coefficients` value to find the amplitude of.
        plot_type: Select the `real`, `imag`, `abs` or `sum` value of the field.
        upsample: Whether to upsample the output.

    Returns:
        The maximum amplitude value.
    """
    # compute inverse transform
    if hasattr(function, "slepian"):
        field = sleplet.slepian_methods.slepian_inverse(
            function.coefficients,
            function.L,
            function.slepian,
        )
    else:
        field = ssht.inverse(
            function.coefficients,
            function.L,
            Method=sleplet._vars.SAMPLING_SCHEME,
        )

    # find resolution of final plot for boosting if necessary
    resolution = calc_plot_resolution(function.L) if upsample else function.L

    # boost field to match final plot
    boosted_field = _boost_field(
        field,
        function.L,
        resolution,
        reality=function.reality,
        spin=function.spin,
        upsample=upsample,
    )

    # find maximum absolute value for given plot type
    return np.abs(_create_plot_type(boosted_field, plot_type)).max()


def _create_plot_type(
    field: npt.NDArray[np.complex128 | np.float64],
    plot_type: str,
) -> npt.NDArray[np.float64]:
    """Get the given plot type of the field."""
    msg = f"plotting type: '{plot_type}'"
    _logger.info(msg)
    plot_dict = {
        "abs": np.abs(field),
        "imag": field.imag,
        "real": field.real,
        "sum": field.real + field.imag,
    }
    return plot_dict[plot_type]


def _set_outside_region_to_minimum(
    f_plot: npt.NDArray[np.float64],
    L: int,
    region: sleplet.slepian.region.Region,
) -> npt.NDArray[np.float64]:
    """
    For the Slepian region set the outside area to negative infinity
    hence it is clear we are only interested in the coloured region.
    """
    # create mask of interest
    mask = sleplet._mask_methods.create_mask_region(L, region)

    # adapt for closed plot
    _, n_phi = ssht.sample_shape(L, Method=sleplet._vars.SAMPLING_SCHEME)
    closed_mask = np.insert(mask, n_phi, mask[:, 0], axis=1)

    # set values outside mask to negative infinity
    return np.where(closed_mask, f_plot, sleplet._vars.SPHERE_UNSEEN)


def _normalise_function(
    f: npt.NDArray[np.float64],
    *,
    normalise: bool,
) -> npt.NDArray[np.float64]:
    """Normalise function between 0 and 1 for visualisation."""
    if not normalise:
        return f
    if (f == 0).all():
        # if all 0, set to 0
        return f + 0.5
    # if all non-zero, set to 1 otherwise scale from [0, 1]
    return f / f.max() if np.allclose(f, f.max()) else (f - f.min()) / f.ptp()


def _boost_field(  # noqa: PLR0913
    field: npt.NDArray[np.complex128 | np.float64],
    L: int,
    resolution: int,
    *,
    reality: bool,
    spin: int,
    upsample: bool,
) -> npt.NDArray[np.complex128 | np.float64]:
    """Inverts and then boosts the field before plotting."""
    if not upsample:
        return field
    flm = ssht.forward(
        field,
        L,
        Method=sleplet._vars.SAMPLING_SCHEME,
        Reality=reality,
        Spin=spin,
    )
    return sleplet.harmonic_methods.invert_flm_boosted(
        flm,
        L,
        resolution,
        reality=reality,
        spin=spin,
    )


def compute_amplitude_for_noisy_mesh_plots(
    f: sleplet.meshes.mesh_coefficients.MeshCoefficients,
) -> float | None:
    """
    For the noised mesh plots fix the amplitude.

    Args:
        f: The mesh field value.

    Returns:
        The noise adjusted maximum amplitude.
    """
    return (
        np.abs(_coefficients_to_field_mesh(f, f._unnoised_coefficients)).max()
        if f._unnoised_coefficients is not None
        else None
    )


def _coefficients_to_field_mesh(
    f: sleplet.meshes.mesh_coefficients.MeshCoefficients,
    coefficients: npt.NDArray[np.complex128 | np.float64],
) -> npt.NDArray[np.complex128 | np.float64]:
    """Compute the field over the whole mesh from the harmonic/Slepian coefficients."""
    return (
        sleplet.slepian_methods.slepian_mesh_inverse(f.mesh_slepian, coefficients)
        if hasattr(f, "mesh_slepian")
        else sleplet.harmonic_methods.mesh_inverse(f.mesh, coefficients)
    )


def compute_amplitude_for_noisy_sphere_plots(
    f: sleplet.functions.coefficients.Coefficients,
) -> float | None:
    """
    For the noised sphere plots fix the amplitude.

    Args:
        f: The spherical field value.

    Returns:
        The noise adjusted maximum amplitude value.
    """
    return (
        np.abs(_coefficients_to_field_sphere(f, f._unnoised_coefficients)).max()
        if f._unnoised_coefficients is not None
        else None
    )


def _coefficients_to_field_sphere(
    f: sleplet.functions.coefficients.Coefficients,
    coefficients: npt.NDArray[np.complex128 | np.float64],
) -> npt.NDArray[np.complex128 | np.float64]:
    """Compute the field over the samples from the harmonic/Slepian coefficients."""
    return (
        sleplet.slepian_methods.slepian_inverse(coefficients, f.L, f.slepian)
        if hasattr(f, "slepian")
        else ssht.inverse(
            coefficients,
            f.L,
            Method=sleplet._vars.SAMPLING_SCHEME,
            Reality=f.reality,
            Spin=f.spin,
        )
    )



"""Contains the `Region` class."""

import logging

import pydantic
import typing_extensions

import sleplet._bool_methods
import sleplet._string_methods
import sleplet._validation
import sleplet._vars

_logger = logging.getLogger(__name__)


@pydantic.dataclasses.dataclass(config=sleplet._validation.validation, kw_only=True)
class Region:
    """Identify and create the appropriate Slepian region for the sphere."""

    gap: bool = False
    """Whether to enable a double ended polar cap, set by the `POLAR_GAP`
    environment variable. Only relevant if `theta_max` is not `180` and the
    other angles are at their default values."""
    mask_name: str = ""
    """The name of the mask of the arbitrary region, set by the `SLEPIAN_MASK`
    environment variable. Current options are `africa` and `south_america`."""
    phi_max: float = sleplet._vars.PHI_MAX_DEFAULT
    """For a limited latitude longitude region, set by the `PHI_MAX` environment
    variable."""
    phi_min: float = sleplet._vars.PHI_MIN_DEFAULT
    """For a limited latitude longitude region, set by the `PHI_MIN` environment
    variable."""
    theta_max: float = sleplet._vars.THETA_MAX_DEFAULT
    """Set by the `THETA_MAX` environment variable. When set without the
    other angles it controls a polar cap region. When in conjunction with the
    others it is for a limited latitude longitude region."""
    theta_min: float = sleplet._vars.THETA_MIN_DEFAULT
    """For a limited latitude longitude region, set by the `THETA_MIN` environment
    variable."""
    _name_ending: str = pydantic.Field(default="", init_var=False, repr=False)
    _region_type: str = pydantic.Field(default="", init_var=False, repr=False)

    def __post_init__(self: typing_extensions.Self) -> None:
        self._identify_region()

    def _identify_region(self: typing_extensions.Self) -> None:
        """Identify region type based on the angle inputs or a mask name."""
        msg = (
            "Slepian region values detected: "
            f"POLAR_GAP={self.gap}, "
            f"THETA_MAX={self.theta_max}, "
            f"THETA_MIN={self.theta_min}, "
            f"PHI_MAX={self.phi_max}, "
            f"PHI_MIN={self.phi_min}, "
            f"SLEPIAN_MASK={self.mask_name}.",
        )
        _logger.info(msg)
        if sleplet._bool_methods.is_polar_cap(
            self.phi_min,
            self.phi_max,
            self.theta_min,
            self.theta_max,
        ):
            self._region_type = "polar"
            self._name_ending = (
                f"polar{'_gap' if self.gap else ''}"
                f"{sleplet._string_methods.angle_as_degree(self.theta_max)}"
            )

        elif sleplet._bool_methods.is_limited_lat_lon(
            self.phi_min,
            self.phi_max,
            self.theta_min,
            self.theta_max,
        ):
            self._region_type = "lim_lat_lon"
            self._name_ending = (
                f"theta{sleplet._string_methods.angle_as_degree(self.theta_min)}"
                f"-{sleplet._string_methods.angle_as_degree(self.theta_max)}"
                f"_phi{sleplet._string_methods.angle_as_degree(self.phi_min)}"
                f"-{sleplet._string_methods.angle_as_degree(self.phi_max)}"
            )

        elif self.mask_name:
            self._region_type = "arbitrary"
            self._name_ending = self.mask_name

        else:
            msg = (
                "need to specify either a polar cap, a limited latitude "
                "longitude region, or a file with a mask",
            )
            raise AttributeError(msg)

    @pydantic.field_validator("phi_max")
    def _check_phi_max(
        cls,  # noqa: ANN101
        v: float,
    ) -> float:
        if v < sleplet._vars.PHI_MIN_DEFAULT:
            msg = "phi_max cannot be negative"
            raise ValueError(msg)
        if v > sleplet._vars.PHI_MAX_DEFAULT:
            msg = (
                f"phi_max cannot be greater than "
                f"{sleplet._string_methods.multiples_of_pi(sleplet._vars.PHI_MAX_DEFAULT)}"
            )
            raise ValueError(msg)
        return v

    @pydantic.field_validator("phi_min")
    def _check_phi_min(
        cls,  # noqa: ANN101
        v: float,
    ) -> float:
        if v < sleplet._vars.PHI_MIN_DEFAULT:
            msg = "phi_min cannot be negative"
            raise ValueError(msg)
        if v > sleplet._vars.PHI_MAX_DEFAULT:
            msg = (
                f"phi_min cannot be greater than "
                f"{sleplet._string_methods.multiples_of_pi(sleplet._vars.PHI_MAX_DEFAULT)}"
            )
            raise ValueError(msg)
        return v

    @pydantic.field_validator("theta_max")
    def _check_theta_max(
        cls,  # noqa: ANN101
        v: float,
    ) -> float:
        if v < sleplet._vars.THETA_MIN_DEFAULT:
            msg = "theta_max cannot be negative"
            raise ValueError(msg)
        if v > sleplet._vars.THETA_MAX_DEFAULT:
            msg = (
                "theta_max cannot be greater than "
                f"{sleplet._string_methods.multiples_of_pi(sleplet._vars.THETA_MAX_DEFAULT)}"
            )
            raise ValueError(msg)
        return v

    @pydantic.field_validator("theta_min")
    def _check_theta_min(
        cls,  # noqa: ANN101
        v: float,
    ) -> float:
        if v < sleplet._vars.THETA_MIN_DEFAULT:
            msg = "theta_min cannot be negative"
            raise ValueError(msg)
        if v > sleplet._vars.THETA_MAX_DEFAULT:
            msg = (
                "theta_min cannot be greater than "
                f"{sleplet._string_methods.multiples_of_pi(sleplet._vars.THETA_MAX_DEFAULT)}"
            )
            raise ValueError(msg)
        return v



import numpy as np
import numpy.typing as npt
import scipy.io as sio

import pyssht as ssht

import sleplet._data.setup_pooch
import sleplet._smoothing


def create_flm(L: int, *, smoothing: int | None = None) -> npt.NDArray[np.complex128]:
    """Create the flm for the whole Earth."""
    # load in data
    flm = _load_flm()

    # fill in negative m components so as to avoid confusion with zero values
    for ell in range(1, L):
        for m in range(1, ell + 1):
            ind_pm = ssht.elm2ind(ell, m)
            ind_nm = ssht.elm2ind(ell, -m)
            flm_pm = flm[ind_pm]
            flm[ind_nm] = (-1) ** m * flm_pm.conj()

    # don't take the full L, invert dataset as Earth backwards
    flm = flm[: L**2].conj()

    if isinstance(smoothing, int):
        flm = sleplet._smoothing.apply_gaussian_smoothing(flm, L, smoothing)
    return flm


def _load_flm() -> npt.NDArray[np.complex128]:
    """Load coefficients from file."""
    mat_contents = sio.loadmat(
        sleplet._data.setup_pooch.find_on_pooch_then_local(
            "EGM2008_Topography_flms_L2190.mat",
        ),
    )
    return np.ascontiguousarray(mat_contents["flm"][:, 0])



"""Methods to work with Slepian coefficients."""

import logging

import numpy as np
import numpy.typing as npt

import pyssht as ssht

import sleplet._vars
import sleplet.harmonic_methods
import sleplet.meshes._mesh_slepian_decomposition
import sleplet.meshes.mesh_slepian
import sleplet.slepian._slepian_decomposition
import sleplet.slepian.slepian_arbitrary
import sleplet.slepian.slepian_limit_lat_lon
import sleplet.slepian.slepian_polar_cap
from sleplet.slepian.region import Region
from sleplet.slepian.slepian_functions import SlepianFunctions

_logger = logging.getLogger(__name__)


def choose_slepian_method(
    L: int,
    region: Region,
) -> SlepianFunctions:
    """
    Initialise Slepian object depending on input.

    Args:
        L: The spherical harmonic bandlimit.
        region: The Slepian region.

    Raises:
        ValueError: Invalid `region_type`, likely cause is invalid `mask_name`.

    Returns:
        The given Slepian object.
    """
    match region._region_type:
        case "polar":
            _logger.info("polar cap region detected")
            return sleplet.slepian.slepian_polar_cap.SlepianPolarCap(
                L,
                region.theta_max,
                gap=region.gap,
            )

        case "lim_lat_lon":
            _logger.info("limited latitude longitude region detected")
            return sleplet.slepian.slepian_limit_lat_lon.SlepianLimitLatLon(
                L,
                theta_min=region.theta_min,
                theta_max=region.theta_max,
                phi_min=region.phi_min,
                phi_max=region.phi_max,
            )

        case "arbitrary":
            _logger.info("mask specified in file detected")
            return sleplet.slepian.slepian_arbitrary.SlepianArbitrary(
                L,
                region.mask_name,
            )

        case _:
            msg = f"{region._region_type} is an invalid region type"
            raise ValueError(msg)


def slepian_inverse(
    f_p: npt.NDArray[np.complex128 | np.float64],
    L: int,
    slepian: SlepianFunctions,
) -> npt.NDArray[np.complex128]:
    """
    Compute the Slepian inverse transform up to the Shannon number.

    Args:
        f_p: The Slepian coefficients.
        L: The spherical harmonic bandlimit.
        slepian: The given Slepian object.

    Returns:
        The values on the sphere in pixel space.
    """
    f_p_reshape = f_p[: slepian.N, np.newaxis, np.newaxis]
    s_p = compute_s_p_omega(L, slepian)
    return (f_p_reshape * s_p).sum(axis=0)


def slepian_forward(  # noqa: PLR0913
    L: int,
    slepian: SlepianFunctions,
    *,
    f: npt.NDArray[np.complex128] | None = None,
    flm: npt.NDArray[np.complex128 | np.float64] | None = None,
    mask: npt.NDArray[np.float64] | None = None,
    n_coeffs: int | None = None,
) -> npt.NDArray[np.complex128]:
    """
    Compute the Slepian forward transform for all coefficients.

    Args:
        L: The spherical harmonic bandlimit.
        slepian: The given Slepian object.
        f: The field value.
        flm: The spherical harmonic coefficients.
        mask: A boolean mask of the Slepian region.
        n_coeffs: The number of Slepian coefficients to use.

    Returns:
        The Slepian coefficients of the inputs.
    """
    sd = sleplet.slepian._slepian_decomposition.SlepianDecomposition(
        L,
        slepian,
        f=f,
        flm=flm,
        mask=mask,
    )
    n_coeffs = slepian.N if n_coeffs is None else n_coeffs
    return sd.decompose_all(n_coeffs)


def compute_s_p_omega(
    L: int,
    slepian: SlepianFunctions,
) -> npt.NDArray[np.complex128]:
    r"""
    Compute \(S_{p}(\omega)\) for a given region.

    Args:
        L: The spherical harmonic bandlimit.
        slepian: The given Slepian object.

    Returns:
        The complex \(S_{p}(\omega)\) values.
    """
    n_theta, n_phi = ssht.sample_shape(L, Method=sleplet._vars.SAMPLING_SCHEME)
    sp = np.zeros((slepian.N, n_theta, n_phi), dtype=np.complex128)
    for p in range(slepian.N):
        if p % L == 0:
            msg = f"compute Sp(omega) p={p+1}/{slepian.N}"
            _logger.info(msg)
        sp[p] = ssht.inverse(
            slepian.eigenvectors[p],
            L,
            Method=sleplet._vars.SAMPLING_SCHEME,
        )
    return sp


def _compute_s_p_omega_prime(
    L: int,
    alpha: float,
    beta: float,
    slepian: SlepianFunctions,
) -> npt.NDArray[np.complex128]:
    """Pick out the desired angle from Sp(omega)."""
    sp_omega = compute_s_p_omega(L, slepian)
    p = ssht.theta_to_index(beta, L, Method=sleplet._vars.SAMPLING_SCHEME)
    q = ssht.phi_to_index(alpha, L, Method=sleplet._vars.SAMPLING_SCHEME)
    sp_omega_prime = sp_omega[:, p, q]
    # pad with zeros so it has the expected shape
    boost = L**2 - slepian.N
    return sleplet.harmonic_methods._boost_coefficient_resolution(sp_omega_prime, boost)


def slepian_mesh_forward(
    mesh_slepian: "sleplet.meshes.mesh_slepian.MeshSlepian",
    *,
    u: npt.NDArray[np.complex128 | np.float64] | None = None,
    u_i: npt.NDArray[np.complex128 | np.float64] | None = None,
    mask: bool = False,
    n_coeffs: int | None = None,
) -> npt.NDArray[np.float64]:
    """
    Compute the Slepian forward transform for all coefficients on the mesh.

    Args:
        mesh_slepian: The Slepian mesh object containing the eigensolutions.
        u: The signal field value on the mesh.
        u_i: The Fourier coefficients of the mesh.
        mask: Whether to use the mask to compute the coefficients.
        n_coeffs: The number of Slepian coefficients to use.

    Returns:
        The Slepian coefficients on the mesh.
    """
    sd = sleplet.meshes._mesh_slepian_decomposition.MeshSlepianDecomposition(
        mesh_slepian,
        u=u,
        u_i=u_i,
        mask=mask,
    )
    n_coeffs = mesh_slepian.N if n_coeffs is None else n_coeffs
    return sd.decompose_all(n_coeffs)


def slepian_mesh_inverse(
    mesh_slepian: "sleplet.meshes.mesh_slepian.MeshSlepian",
    f_p: npt.NDArray[np.complex128 | np.float64],
) -> npt.NDArray[np.complex128 | np.float64]:
    """
    Compute the Slepian inverse transform on the mesh up to the Shannon number.

    Args:
        mesh_slepian: The Slepian mesh object containing the eigensolutions.
        f_p: The Slepian wavelet coefficients.

    Returns:
        The value of a function on the mesh in pixel space.
    """
    f_p_reshape = f_p[: mesh_slepian.N, np.newaxis]
    s_p = _compute_mesh_s_p_pixel(mesh_slepian)
    return (f_p_reshape * s_p).sum(axis=0)


def _compute_mesh_s_p_pixel(
    mesh_slepian: "sleplet.meshes.mesh_slepian.MeshSlepian",
) -> npt.NDArray[np.float64]:
    """Calculate Sp(omega) for a given region."""
    sp = np.zeros((mesh_slepian.N, mesh_slepian.mesh.vertices.shape[0]))
    for p in range(mesh_slepian.N):
        sp[p] = sleplet.harmonic_methods.mesh_inverse(
            mesh_slepian.mesh,
            mesh_slepian.slepian_functions[p],
        )
    return sp



"""Contains the abstract `MeshCoefficients` class."""

import abc
import dataclasses

import numpy as np
import numpy.typing as npt
import pydantic
import typing_extensions

import sleplet._mask_methods
import sleplet._string_methods
import sleplet._validation
import sleplet._vars
from sleplet.meshes.mesh import Mesh

_COEFFICIENTS_TO_NOT_MASK: str = "slepian"


@pydantic.dataclasses.dataclass(config=sleplet._validation.validation)
class MeshCoefficients:
    """Abstract parent class to handle Fourier/Slepian coefficients on the mesh."""

    mesh: Mesh
    """A mesh object."""
    _: dataclasses.KW_ONLY
    extra_args: list[int] | None = None
    """Control the extra arguments for the given set of mesh
    coefficients. Only to be set by the `mesh` CLI."""
    noise: float | None = None
    """How much to noise the data."""
    region: bool = False
    """Whether to set a region or not, used in the Slepian case."""
    _unnoised_coefficients: npt.NDArray[np.complex128 | np.float64] | None = (
        pydantic.Field(
            default=None,
            init_var=False,
            repr=False,
        )
    )
    coefficients: npt.NDArray[np.complex128 | np.float64] = pydantic.Field(
        default_factory=lambda: np.empty(0),
        init_var=False,
        repr=False,
    )
    name: str = pydantic.Field(default="", init_var=False, repr=False)
    snr: float | None = pydantic.Field(default=None, init_var=False, repr=False)
    wavelet_coefficients: npt.NDArray[np.complex128 | np.float64] = pydantic.Field(
        default_factory=lambda: np.empty(0),
        init_var=False,
        repr=False,
    )
    wavelets: npt.NDArray[np.float64] = pydantic.Field(
        default_factory=lambda: np.empty(0),
        init_var=False,
        repr=False,
    )

    def __post_init__(self: typing_extensions.Self) -> None:
        self._setup_args()
        self.name = self._create_name()
        self.coefficients = self._create_coefficients()
        self._add_details_to_name()
        self._unnoised_coefficients, self.snr = self._add_noise_to_signal()

    def _add_details_to_name(self: typing_extensions.Self) -> None:
        """Add region to the name if present if not a Slepian function."""
        if self.region and "slepian" not in self.mesh.name:
            self.name += "_region"
        if self.noise is not None:
            self.name += f"{sleplet._string_methods.filename_args(self.noise, 'noise')}"
        if self.mesh.zoom:
            self.name += "_zoom"

    @abc.abstractmethod
    def _add_noise_to_signal(
        self: typing_extensions.Self,
    ) -> tuple[npt.NDArray[np.complex128 | np.float64] | None, float | None]:
        """Add Gaussian white noise to the signal."""
        raise NotImplementedError

    @abc.abstractmethod
    def _create_coefficients(
        self: typing_extensions.Self,
    ) -> npt.NDArray[np.complex128 | np.float64]:
        """Create the flm on the north pole."""
        raise NotImplementedError

    @abc.abstractmethod
    def _create_name(self: typing_extensions.Self) -> str:
        """Create the name of the function."""
        raise NotImplementedError

    @abc.abstractmethod
    def _setup_args(self: typing_extensions.Self) -> None:
        """Initialise function specific args either default value or user input."""
        raise NotImplementedError



"""Contains the `MeshSlepianField` class."""

import numpy as np
import numpy.typing as npt
import pydantic
import typing_extensions

import sleplet._validation
import sleplet.meshes.mesh_field
import sleplet.slepian_methods
from sleplet.meshes.mesh_slepian_coefficients import MeshSlepianCoefficients


@pydantic.dataclasses.dataclass(config=sleplet._validation.validation)
class MeshSlepianField(MeshSlepianCoefficients):
    """
    Create a field on a given mesh computed from a Slepian region of the mesh.
    The default field is the per-vertex normals of the mesh.
    """

    def __post_init__(self: typing_extensions.Self) -> None:
        super().__post_init__()

    def _create_coefficients(
        self: typing_extensions.Self,
    ) -> npt.NDArray[np.complex128 | np.float64]:
        """Compute field on the vertices of the mesh."""
        mf = sleplet.meshes.mesh_field.MeshField(
            self.mesh,
            region=True,
        )
        return sleplet.slepian_methods.slepian_mesh_forward(
            self.mesh_slepian,
            u_i=mf.coefficients,
        )

    def _create_name(self: typing_extensions.Self) -> str:
        return f"slepian_{self.mesh.name}_field"

    def _setup_args(self: typing_extensions.Self) -> None:
        if isinstance(self.extra_args, list):
            msg = f"{self.__class__.__name__} does not support extra arguments"
            raise TypeError(msg)



"""Contains the `Africa` class."""

import numpy as np
import numpy.typing as npt
import pydantic
import typing_extensions

import pyssht as ssht

import sleplet._data.create_earth_flm
import sleplet._data.setup_pooch
import sleplet._mask_methods
import sleplet._string_methods
import sleplet._validation
import sleplet._vars
import sleplet.harmonic_methods
from sleplet.functions.flm import Flm


@pydantic.dataclasses.dataclass(config=sleplet._validation.validation)
class Africa(Flm):
    """
    Create a topographic map of the Earth setting everything outide of the
    Africa region to zero.
    """

    def __post_init__(self: typing_extensions.Self) -> None:
        super().__post_init__()

    def _create_coefficients(
        self: typing_extensions.Self,
    ) -> npt.NDArray[np.complex128 | np.float64]:
        return sleplet.harmonic_methods._ensure_f_bandlimited(
            self._grid_fun,
            self.L,
            reality=self.reality,
            spin=self.spin,
        )

    def _create_name(self: typing_extensions.Self) -> str:
        return sleplet._string_methods._convert_camel_case_to_snake_case(
            self.__class__.__name__,
        )

    def _set_reality(self: typing_extensions.Self) -> bool:
        return True

    def _set_spin(self: typing_extensions.Self) -> int:
        return 0

    def _setup_args(self: typing_extensions.Self) -> None:
        if isinstance(self.extra_args, list):
            msg = f"{self.__class__.__name__} does not support extra arguments"
            raise TypeError(msg)

    def _grid_fun(
        self: typing_extensions.Self,
        theta: npt.NDArray[np.float64],  # noqa: ARG002
        phi: npt.NDArray[np.float64],  # noqa: ARG002
    ) -> npt.NDArray[np.float64]:
        """Define the function on the grid."""
        earth_flm = sleplet._data.create_earth_flm.create_flm(
            self.L,
            smoothing=self.smoothing,
        )
        rot_flm = sleplet.harmonic_methods.rotate_earth_to_africa(earth_flm, self.L)
        earth_f = ssht.inverse(
            rot_flm,
            self.L,
            Method=sleplet._vars.SAMPLING_SCHEME,
            Reality=self.reality,
        )
        mask_name = f"{self.name}_L{self.L}.npy"
        mask_location = sleplet._data.setup_pooch.find_on_pooch_then_local(
            f"slepian_masks_{mask_name}",
        )
        mask = (
            sleplet._mask_methods.create_mask(self.L, mask_name)
            if mask_location is None
            else np.load(mask_location)
        )
        return np.where(mask, earth_f, 0)



import numpy as np
import numpy.typing as npt
import scipy.io as sio

import pyssht as ssht

import sleplet._data.setup_pooch
import sleplet._vars


def create_flm(L: int) -> npt.NDArray[np.complex128]:
    """Create the flm for the whole CMB."""
    # load in data
    cl = _load_cl()

    # same random seed
    rng = np.random.default_rng(sleplet._vars.RANDOM_SEED)

    # Simulate CMB in harmonic space.
    flm = np.zeros(L**2, dtype=np.complex128)
    for ell in range(2, L):
        sigma = np.sqrt(2 * np.pi / (ell * (ell + 1)) * cl[ell - 2])
        ind = ssht.elm2ind(ell, 0)
        flm[ind] = sigma * rng.standard_normal()
        for m in range(1, ell + 1):
            ind_pm = ssht.elm2ind(ell, m)
            ind_nm = ssht.elm2ind(ell, -m)
            flm[ind_pm] = (
                sigma
                / np.sqrt(2)
                * (rng.standard_normal() + 1j * rng.standard_normal())
            )
            flm[ind_nm] = (-1) ** m * flm[ind_pm].conj()
    return flm


def _load_cl(
    *,
    file_ending: str = "_lcdm_pl_model_wmap7baoh0",
) -> npt.NDArray[np.float64]:
    """
    Pick coefficients from file options are:
    * _lcdm_pl_model_yr1_v1.mat
    * _tt_spectrum_7yr_v4p1.mat
    * _lcdm_pl_model_wmap7baoh0.mat.
    """
    mat_contents = sio.loadmat(
        sleplet._data.setup_pooch.find_on_pooch_then_local(f"wmap{file_ending}.mat"),
    )
    return np.ascontiguousarray(mat_contents["cl"][:, 0])



"""Contains the `NoiseEarth` class."""

import numpy as np
import numpy.typing as npt
import pydantic
import typing_extensions

import sleplet._string_methods
import sleplet._validation
import sleplet.functions.earth
import sleplet.noise
from sleplet.functions.flm import Flm


@pydantic.dataclasses.dataclass(config=sleplet._validation.validation, kw_only=True)
class NoiseEarth(Flm):
    """Create a noised signal of the topographic map of the Earth."""

    SNR: float = 10
    """A parameter which controls the level of signal-to-noise in the noised
    data."""

    def __post_init__(self: typing_extensions.Self) -> None:
        super().__post_init__()

    def _create_coefficients(
        self: typing_extensions.Self,
    ) -> npt.NDArray[np.complex128 | np.float64]:
        earth = sleplet.functions.earth.Earth(self.L, smoothing=self.smoothing)
        noise = sleplet.noise._create_noise(self.L, earth.coefficients, self.SNR)
        sleplet.noise.compute_snr(earth.coefficients, noise, "Harmonic")
        return noise

    def _create_name(self: typing_extensions.Self) -> str:
        return (
            f"{sleplet._string_methods._convert_camel_case_to_snake_case(self.__class__.__name__)}"
            f"{sleplet._string_methods.filename_args(self.SNR, 'snr')}"
        )

    def _set_reality(self: typing_extensions.Self) -> bool:
        return True

    def _set_spin(self: typing_extensions.Self) -> int:
        return 0

    def _setup_args(self: typing_extensions.Self) -> None:
        if isinstance(self.extra_args, list):
            num_args = 1
            if len(self.extra_args) != num_args:
                msg = f"The number of extra arguments should be {num_args}"
                raise ValueError(msg)
            self.SNR = self.extra_args[0]



import sleplet._vars


def is_polar_cap(
    phi_min: float,
    phi_max: float,
    theta_min: float,
    theta_max: float,
) -> bool:
    """Circular mask at the north pole."""
    return (
        phi_min == sleplet._vars.PHI_MIN_DEFAULT
        and phi_max == sleplet._vars.PHI_MAX_DEFAULT
        and theta_min == sleplet._vars.THETA_MIN_DEFAULT
        and theta_max != sleplet._vars.THETA_MAX_DEFAULT
    )


def is_limited_lat_lon(
    phi_min: float,
    phi_max: float,
    theta_min: float,
    theta_max: float,
) -> bool:
    """Define a region by angles, just need one to not be the default."""
    return (
        not is_polar_cap(phi_min, phi_max, theta_min, theta_max)
        and phi_min != sleplet._vars.PHI_MIN_DEFAULT
        or phi_max != sleplet._vars.PHI_MAX_DEFAULT
        or theta_min != sleplet._vars.THETA_MIN_DEFAULT
    )



import logging
import os
import typing

import platformdirs
import pooch

_logger = logging.getLogger(__name__)


_POOCH = None
_ZENODO_DATA_DOI = "10.5281/zenodo.7767698"


def _lazy_load_registry(
    func: typing.Callable[..., os.PathLike[str] | None],
) -> typing.Callable[..., os.PathLike[str] | None]:
    """
    Lazily loads POOCH registry before executing a function.

    Args:
        func: Function to be decorated

    Returns:
        Decorated function
    """

    def wrapper(
        *args: typing.Any,  # noqa: ANN401
        **kwargs: typing.Any,  # noqa: ANN401
    ) -> os.PathLike[str] | None:
        """
        Load the POOCH registry if not already loaded and execute the function.

        Args:
            *args: Positional arguments for the function
            **kwargs: Keyword arguments for the function.

        Returns:
            Result of the decorated function
        """
        global _POOCH  # noqa: PLW0603
        if _POOCH is None:
            _POOCH = pooch.create(
                path=pooch.os_cache("sleplet"),
                base_url=f"doi:{_ZENODO_DATA_DOI}/",
                registry=None,
            )
            _POOCH.load_registry_from_doi()
        return func(*args, **kwargs)

    return wrapper


@_lazy_load_registry
def find_on_pooch_then_local(filename: str) -> os.PathLike[str] | None:
    """
    Find a file on POOCH first and if not look in data folder.

    Args:
        filename: Filename to find

    Returns:
        The sought after file or nothing if not found
    """
    if filename in _POOCH.registry:  # type: ignore[union-attr]
        msg = f"Found {filename} at https://doi.org/{_ZENODO_DATA_DOI}"
        _logger.info(msg)
        return _POOCH.fetch(filename, progressbar=True)  # type: ignore[union-attr]

    if (platformdirs.user_data_path() / filename).exists():
        msg = f"Found {filename} at {platformdirs.user_data_path() / filename}"
        _logger.info(msg)
        return platformdirs.user_data_path() / filename

    msg = f"No {filename} found, calculating..."
    _logger.info(msg)
    return None



"""Methods to perform operations in Fourier space of the sphere or mesh."""

import collections
import typing

import numpy as np
import numpy.typing as npt

import pyssht as ssht

import sleplet._integration_methods
import sleplet._vars
import sleplet.meshes.mesh

_AFRICA_ALPHA = np.deg2rad(44)
_AFRICA_BETA = np.deg2rad(87)
_AFRICA_GAMMA = np.deg2rad(341)
_SOUTH_AMERICA_ALPHA = np.deg2rad(54)
_SOUTH_AMERICA_BETA = np.deg2rad(108)
_SOUTH_AMERICA_GAMMA = np.deg2rad(63)


def _create_spherical_harmonic(L: int, ind: int) -> npt.NDArray[np.complex128]:
    """Create a spherical harmonic in harmonic space for the given index."""
    flm = np.zeros(L**2, dtype=np.complex128)
    flm[ind] = 1
    return flm


def _boost_coefficient_resolution(
    flm: npt.NDArray[typing.Any],
    boost: int,
) -> npt.NDArray[typing.Any]:
    """Calculate a boost in resolution for given flm."""
    return np.pad(flm, (0, boost), "constant")


def invert_flm_boosted(
    flm: npt.NDArray[np.complex128],
    L: int,
    resolution: int,
    *,
    reality: bool = False,
    spin: int = 0,
) -> npt.NDArray[np.complex128 | np.float64]:
    """
    Upsamples the signal and performs the inverse harmonic transform .

    Args:
        flm: The spherical harmonic coefficients.
        L: The spherical harmonic bandlimit.
        resolution: The output resolution of the field values.
        reality: Whether the given spherical signal is real or not.
        spin: The value of the spin.

    Returns:
        The boosted field value.
    """
    boost = resolution**2 - L**2
    flm = _boost_coefficient_resolution(flm, boost)
    return ssht.inverse(
        flm,
        resolution,
        Method=sleplet._vars.SAMPLING_SCHEME,
        Reality=reality,
        Spin=spin,
    )


def _ensure_f_bandlimited(
    grid_fun: collections.abc.Callable[
        [npt.NDArray[np.float64], npt.NDArray[np.float64]],
        npt.NDArray[np.float64],
    ],
    L: int,
    *,
    reality: bool,
    spin: int,
) -> npt.NDArray[np.complex128]:
    """
    If the function created is created in pixel space rather than harmonic
    space then need to transform it into harmonic space first before using it.
    """
    thetas, phis = ssht.sample_positions(
        L,
        Grid=True,
        Method=sleplet._vars.SAMPLING_SCHEME,
    )
    f = grid_fun(thetas, phis)
    return ssht.forward(
        f,
        L,
        Method=sleplet._vars.SAMPLING_SCHEME,
        Reality=reality,
        Spin=spin,
    )


def _create_emm_vector(L: int) -> npt.NDArray[np.float64]:
    """Create vector of m values for a given L."""
    emm = np.zeros(2 * L * 2 * L)
    k = 0

    for ell in range(2 * L):
        M = 2 * ell + 1
        emm[k : k + M] = np.arange(-ell, ell + 1)
        k += M
    return emm


def compute_random_signal(
    L: int,
    rng: np.random.Generator,
    *,
    var_signal: float,
) -> npt.NDArray[np.complex128]:
    """
    Generate a normally distributed random signal of a
    complex signal with mean `0` and variance `1`.

    Args:
        L: The spherical harmonic bandlimit.
        rng: The random number generator object.
        var_signal: The variance of the signal.

    Returns:
        The coefficients of a random signal on the sphere.
    """
    return np.sqrt(var_signal / 2) * (
        rng.standard_normal(L**2) + 1j * rng.standard_normal(L**2)
    )


def mesh_forward(
    mesh: "sleplet.meshes.mesh.Mesh",
    u: npt.NDArray[np.complex128 | np.float64],
) -> npt.NDArray[np.float64]:
    """
    Compute the mesh forward transform from pixel space to Fourier space.

    Args:
        mesh: The given mesh object.
        u: The signal field value on the mesh.

    Returns:
        The basis functions of the mesh in Fourier space.
    """
    u_i = np.zeros(mesh.mesh_eigenvalues.shape[0])
    for i, phi_i in enumerate(mesh.basis_functions):
        u_i[i] = sleplet._integration_methods.integrate_whole_mesh(
            mesh.vertices,
            mesh.faces,
            u,
            phi_i,
        )
    return u_i


def mesh_inverse(
    mesh: "sleplet.meshes.mesh.Mesh",
    u_i: npt.NDArray[np.complex128 | np.float64],
) -> npt.NDArray[np.complex128 | np.float64]:
    """
    Compute the mesh inverse transform from Fourier space to pixel space.

    Args:
        mesh: The given mesh object.
        u_i: The Fourier coefficients on the mesh.

    Returns:
        The values on the mesh in pixel space.
    """
    return (u_i[:, np.newaxis] * mesh.basis_functions).sum(axis=0)


def rotate_earth_to_south_america(
    earth_flm: npt.NDArray[np.complex128 | np.float64],
    L: int,
) -> npt.NDArray[np.complex128]:
    """
    Rotates the harmonic coefficients of the Earth to a view centered on South America.

    Args:
        earth_flm: The spherical harmonic coefficients of the Earth.
        L: The spherical harmonic bandlimit.

    Returns:
        The spherical harmonic coefficients of the Earth centered on South America.
    """
    return ssht.rotate_flms(
        earth_flm,
        _SOUTH_AMERICA_ALPHA,
        _SOUTH_AMERICA_BETA,
        _SOUTH_AMERICA_GAMMA,
        L,
    )


def rotate_earth_to_africa(
    earth_flm: npt.NDArray[np.complex128 | np.float64],
    L: int,
) -> npt.NDArray[np.complex128]:
    """
    Rotates the harmonic coefficients of the Earth to a view centered on Africa.

    Args:
        earth_flm: The spherical harmonic coefficients of the Earth.
        L: The spherical harmonic bandlimit.

    Returns:
        The spherical harmonic coefficients of the Earth centered on Africa.
    """
    return ssht.rotate_flms(earth_flm, _AFRICA_ALPHA, _AFRICA_BETA, _AFRICA_GAMMA, L)



import logging
import os
import typing

import numpy as np
import numpy.typing as npt
import platformdirs

import pyssht as ssht

import sleplet._data.create_earth_flm
import sleplet._data.setup_pooch
import sleplet._vars
import sleplet.harmonic_methods
import sleplet.meshes.mesh
import sleplet.slepian.region

_logger = logging.getLogger(__name__)

_AFRICA_RANGE = np.deg2rad(41)
_SOUTH_AMERICA_RANGE = np.deg2rad(40)


def create_mask_region(
    L: int,
    region: "sleplet.slepian.region.Region",
) -> npt.NDArray[np.float64]:
    """
    Create a mask of a region of interested, the output will be based
    on the value of the provided L. The mask could be either:
    * polar cap - if theta_max provided
    * limited latitude longitude - if one of theta_min, theta_max,
                                   phi_min or phi_max is provided
    * arbitrary - just checks the shape of the input mask.
    """
    thetas, phis = ssht.sample_positions(
        L,
        Grid=True,
        Method=sleplet._vars.SAMPLING_SCHEME,
    )

    match region._region_type:
        case "arbitrary":
            _logger.info("loading and checking shape of provided mask")
            name = f"{region.mask_name}_L{L}.npy"
            mask = _load_mask(L, name)
            assert mask.shape == thetas.shape, (  # noqa: S101
                f"mask {name} has shape {mask.shape} which does not match "
                f"the provided L={L}, the shape should be {thetas.shape}"
            )

        case "lim_lat_lon":
            _logger.info("creating limited latitude longitude mask")
            mask = (
                (thetas >= region.theta_min)
                & (thetas <= region.theta_max)
                & (phis >= region.phi_min)
                & (phis <= region.phi_max)
            )

        case "polar":
            _logger.info("creating polar cap mask")
            mask = thetas <= region.theta_max
            if region.gap:
                _logger.info("creating polar gap mask")
                mask += thetas >= np.pi - region.theta_max
    return mask


def _load_mask(L: int, mask_name: str) -> npt.NDArray[np.float64]:
    """Attempt to read the mask from the config file."""
    mask = sleplet._data.setup_pooch.find_on_pooch_then_local(
        f"slepian_masks_{mask_name}",
    )
    return create_mask(L, mask_name) if mask is None else np.load(mask)


def ensure_masked_flm_bandlimited(
    flm: npt.NDArray[np.complex128],
    L: int,
    region: "sleplet.slepian.region.Region",
    *,
    reality: bool,
    spin: int,
) -> npt.NDArray[np.complex128]:
    """Ensure the coefficients is bandlimited for a given region."""
    field = ssht.inverse(
        flm,
        L,
        Method=sleplet._vars.SAMPLING_SCHEME,
        Reality=reality,
        Spin=spin,
    )
    mask = create_mask_region(L, region)
    field = np.where(mask, field, 0)
    return ssht.forward(
        field,
        L,
        Method=sleplet._vars.SAMPLING_SCHEME,
        Reality=reality,
        Spin=spin,
    )


def create_default_region() -> "sleplet.slepian.region.Region":
    """Create default region."""
    return sleplet.slepian.region.Region(
        gap=os.getenv("POLAR_GAP", "False").lower() == "true",
        mask_name=os.getenv("SLEPIAN_MASK", "south_america"),
        phi_max=np.deg2rad(int(os.getenv("PHI_MAX", "360"))),
        phi_min=np.deg2rad(int(os.getenv("PHI_MIN", "0"))),
        theta_max=np.deg2rad(int(os.getenv("THETA_MAX", "180"))),
        theta_min=np.deg2rad(int(os.getenv("THETA_MIN", "0"))),
    )


def create_mesh_region(
    mesh_config: dict[str, typing.Any],
    vertices: npt.NDArray[np.float64],
) -> npt.NDArray[np.bool_]:
    """Create a boolean region for the given mesh."""
    return (
        (vertices[:, 0] >= mesh_config["XMIN"])
        & (vertices[:, 0] <= mesh_config["XMAX"])
        & (vertices[:, 1] >= mesh_config["YMIN"])
        & (vertices[:, 1] <= mesh_config["YMAX"])
        & (vertices[:, 2] >= mesh_config["ZMIN"])
        & (vertices[:, 2] <= mesh_config["ZMAX"])
    )


def ensure_masked_bandlimit_mesh_signal(
    mesh: "sleplet.meshes.mesh.Mesh",
    u_i: npt.NDArray[np.complex128 | np.float64],
) -> npt.NDArray[np.float64]:
    """Ensure that signal in pixel space is bandlimited."""
    field = sleplet.harmonic_methods.mesh_inverse(mesh, u_i)
    masked_field = np.where(mesh.mesh_region, field, 0)
    return sleplet.harmonic_methods.mesh_forward(mesh, masked_field)


def convert_region_on_vertices_to_faces(
    mesh: "sleplet.meshes.mesh.Mesh",
) -> npt.NDArray[np.float64]:
    """Convert the region on vertices to faces."""
    region_reshape = np.argwhere(mesh.mesh_region).reshape(-1)
    faces_in_region = np.isin(mesh.faces, region_reshape).all(axis=1)
    region_on_faces = np.zeros(mesh.faces.shape[0])
    region_on_faces[faces_in_region] = 1
    return region_on_faces


def _create_africa_mask(
    L: int,
    earth_flm: npt.NDArray[np.complex128],
) -> npt.NDArray[np.float64]:
    """Create the Africa region mask."""
    rot_flm = sleplet.harmonic_methods.rotate_earth_to_africa(earth_flm, L)
    earth_f = ssht.inverse(
        rot_flm,
        L,
        Method=sleplet._vars.SAMPLING_SCHEME,
        Reality=True,
    )
    thetas, _ = ssht.sample_positions(
        L,
        Grid=True,
        Method=sleplet._vars.SAMPLING_SCHEME,
    )
    return (thetas <= _AFRICA_RANGE) & (earth_f >= 0)


def _create_south_america_mask(
    L: int,
    earth_flm: npt.NDArray[np.complex128],
) -> npt.NDArray[np.float64]:
    """Create the Africa region mask."""
    rot_flm = sleplet.harmonic_methods.rotate_earth_to_south_america(earth_flm, L)
    earth_f = ssht.inverse(
        rot_flm,
        L,
        Method=sleplet._vars.SAMPLING_SCHEME,
        Reality=True,
    )
    thetas, _ = ssht.sample_positions(
        L,
        Grid=True,
        Method=sleplet._vars.SAMPLING_SCHEME,
    )
    return (thetas <= _SOUTH_AMERICA_RANGE) & (earth_f >= 0)


def create_mask(L: int, mask_name: str) -> npt.NDArray[np.float64]:
    """Create the South America region mask."""
    earth_flm = sleplet._data.create_earth_flm.create_flm(L)
    if mask_name == f"africa_L{L}.npy":
        mask = _create_africa_mask(L, earth_flm)
    elif mask_name == f"south_america_L{L}.npy":
        mask = _create_south_america_mask(L, earth_flm)
    else:
        msg = f"Mask name {mask_name} not recognised"
        raise ValueError(msg)
    np.save(platformdirs.user_data_path() / f"slepian_masks_{mask_name}", mask)
    return mask



import typing

import plotly.graph_objs as go

_axis = {
    "title": "",
    "showgrid": False,
    "zeroline": False,
    "ticks": "",
    "showticklabels": False,
    "showbackground": False,
}


def create_camera(  # noqa: PLR0913
    x_eye: float,
    y_eye: float,
    z_eye: float,
    zoom: float,
    *,
    x_center: float = 0,
    y_center: float = 0,
    z_center: float = 0,
) -> go.layout.scene.Camera:
    """Create default camera view with a zoom factor."""
    return go.layout.scene.Camera(
        eye=go.layout.scene.camera.Eye(x=x_eye / zoom, y=y_eye / zoom, z=z_eye / zoom),
        center=go.layout.scene.camera.Center(x=x_center, y=y_center, z=z_center),
    )


def create_layout(
    camera: go.layout.scene.Camera,
    *,
    annotations: list[dict[str, float | int]] | None = None,
) -> go.Layout:
    """Create the default plotly layout."""
    return go.Layout(
        scene=go.layout.Scene(
            dragmode="orbit",
            camera=camera,
            xaxis=go.layout.scene.XAxis(_axis),
            yaxis=go.layout.scene.YAxis(_axis),
            zaxis=go.layout.scene.ZAxis(_axis),
            annotations=annotations,
        ),
        margin=go.layout.Margin(l=0, r=0, b=0, t=0),
        paper_bgcolor="rgba(0,0,0,0)",
        plot_bgcolor="rgba(0,0,0,0)",
    )


def create_tick_mark(
    fmin: float,
    fmax: float,
    *,
    amplitude: float | None = None,
) -> float:
    """Create tick mark to use when using a non-normalised plot."""
    return amplitude if amplitude is not None else max(abs(fmin), abs(fmax))


def create_colour_bar(
    tick_mark: float,
    *,
    normalise: bool,
    bar_len: float = 0.94,
    bar_pos: float = 0.97,
    font_size: int = 38,
) -> dict[str, typing.Any]:
    """Create the default plotly colour bar."""
    return {
        "x": bar_pos,
        "len": bar_len,
        "nticks": 2 if normalise else None,
        "tickfont": {"color": "#666666", "size": font_size},
        "tickformat": None if normalise else "+.1e",
        "tick0": None if normalise else -tick_mark,
        "dtick": None if normalise else tick_mark,
    }



"""Contains the `SlepianArbitrary` class."""

import concurrent.futures
import logging
import os

import numpy as np
import numpy.linalg as LA  # noqa: N812
import numpy.typing as npt
import platformdirs
import pydantic
import typing_extensions

import pyssht as ssht

import sleplet._array_methods
import sleplet._data.setup_pooch
import sleplet._integration_methods
import sleplet._mask_methods
import sleplet._parallel_methods
import sleplet._slepian_arbitrary_methods
import sleplet._validation
import sleplet.harmonic_methods
import sleplet.slepian.region
from sleplet.slepian.slepian_functions import SlepianFunctions

_logger = logging.getLogger(__name__)

_SAMPLES = 2


@pydantic.dataclasses.dataclass(config=sleplet._validation.validation)
class SlepianArbitrary(SlepianFunctions):
    """Class to create an arbitrary Slepian region on the sphere."""

    mask_name: str
    """The name of the mask of the arbitrary region."""
    _weight: npt.NDArray[np.float64] = pydantic.Field(
        default_factory=lambda: np.empty(0),
        init_var=False,
        repr=False,
    )

    def __post_init__(self: typing_extensions.Self) -> None:
        self._resolution = _SAMPLES * self.L
        super().__post_init__()

    def _create_fn_name(self: typing_extensions.Self) -> str:
        return f"slepian_{self.mask_name}"

    def _create_region(self: typing_extensions.Self) -> "sleplet.slepian.region.Region":
        return sleplet.slepian.region.Region(mask_name=self.mask_name)

    def _create_mask(self: typing_extensions.Self) -> npt.NDArray[np.float64]:
        return sleplet._mask_methods.create_mask_region(self._resolution, self.region)

    def _calculate_area(self: typing_extensions.Self) -> float:
        self._weight = sleplet._integration_methods.calc_integration_weight(
            self._resolution,
        )
        return (self.mask * self._weight).sum()

    def _create_matrix_location(self: typing_extensions.Self) -> str:
        return f"slepian_eigensolutions_D_{self.mask_name}_L{self.L}_N{self.N}"

    def _solve_eigenproblem(
        self: typing_extensions.Self,
    ) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.complex128]]:
        eval_loc = f"{self.matrix_location}_eigenvalues.npy"
        evec_loc = f"{self.matrix_location}_eigenvectors.npy"

        try:
            eigenvalues = np.load(
                sleplet._data.setup_pooch.find_on_pooch_then_local(eval_loc),
            )
            eigenvectors = np.load(
                sleplet._data.setup_pooch.find_on_pooch_then_local(evec_loc),
            )
        except TypeError:
            eigenvalues, eigenvectors = self._solve_D_matrix(eval_loc, evec_loc)
        return eigenvalues, eigenvectors

    def _solve_D_matrix(  # noqa: N802
        self: typing_extensions.Self,
        eval_loc: str,
        evec_loc: str,
    ) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.complex128]]:
        D = self._create_D_matrix()

        # fill in remaining triangle section
        sleplet._array_methods.fill_upper_triangle_of_hermitian_matrix(D)

        # solve eigenproblem
        (
            eigenvalues,
            eigenvectors,
        ) = sleplet._slepian_arbitrary_methods.clean_evals_and_evecs(LA.eigh(D))
        np.save(platformdirs.user_data_path() / eval_loc, eigenvalues)
        np.save(platformdirs.user_data_path() / evec_loc, eigenvectors[: self.N])
        return eigenvalues, eigenvectors

    def _create_D_matrix(  # noqa: N802
        self: typing_extensions.Self,
    ) -> npt.NDArray[np.complex128]:
        """Compute the D matrix in parallel."""
        # create dictionary for the integrals
        self._fields: dict[int, npt.NDArray[np.complex128 | np.float64]] = {}

        # initialise real and imaginary matrices
        D_r = np.zeros((self.L**2, self.L**2))
        D_i = np.zeros((self.L**2, self.L**2))

        D_r_ext, shm_r_ext = sleplet._parallel_methods.create_shared_memory_array(D_r)
        D_i_ext, shm_i_ext = sleplet._parallel_methods.create_shared_memory_array(D_i)

        def func(chunk: list[int]) -> None:
            """Calculate D matrix components for each chunk."""
            (
                D_r_int,
                shm_r_int,
            ) = sleplet._parallel_methods.attach_to_shared_memory_block(D_r, shm_r_ext)
            (
                D_i_int,
                shm_i_int,
            ) = sleplet._parallel_methods.attach_to_shared_memory_block(D_i, shm_i_ext)

            for i in chunk:
                msg = f"start ell: {i}"
                _logger.info(msg)
                self._matrix_helper(D_r_int, D_i_int, i)
                msg = f"finish ell: {i}"
                _logger.info(msg)

            sleplet._parallel_methods.free_shared_memory(shm_r_int, shm_i_int)

        # split up L range to maximise efficiency
        ncpu = int(os.getenv("NCPU", "4"))
        msg = f"Number of CPU={ncpu}"
        _logger.info(msg)
        chunks = sleplet._parallel_methods.split_arr_into_chunks(
            self.L**2,
            ncpu,
        )

        # initialise pool and apply function
        with concurrent.futures.ThreadPoolExecutor(max_workers=ncpu) as e:
            e.map(func, chunks)

        # retrieve from parallel function
        D = D_r_ext + 1j * D_i_ext

        # Free and release the shared memory block at the very end
        sleplet._parallel_methods.free_shared_memory(shm_r_ext, shm_i_ext)
        sleplet._parallel_methods.release_shared_memory(shm_r_ext, shm_i_ext)
        return D

    def _matrix_helper(
        self: typing_extensions.Self,
        D_r: npt.NDArray[np.float64],
        D_i: npt.NDArray[np.float64],
        i: int,
    ) -> None:
        """
        Use in both serial and parallel calculations.

        The hack with splitting into real and imaginary parts
        is not required for the serial case but here for ease
        """
        # fill in diagonal components
        integral = self._integral(i, i)
        D_r[i][i] = integral.real
        D_i[i][i] = integral.imag
        _, m_i = ssht.ind2elm(i)

        for j in range(i + 1, D_r.shape[0]):
            ell_j, m_j = ssht.ind2elm(j)
            # if possible to use previous calculations
            if m_i == 0 and m_j != 0 and ell_j < self.L:
                # if positive m then use conjugate relation
                if m_j > 0:
                    integral = self._integral(j, i)
                    D_r[j][i] = integral.real
                    D_i[j][i] = integral.imag
                    k = ssht.elm2ind(ell_j, -m_j)
                    D_r[k][i] = (-1) ** m_j * D_r[j][i]
                    D_i[k][i] = (-1) ** (m_j + 1) * D_i[j][i]
            else:
                integral = self._integral(j, i)
                D_r[j][i] = integral.real
                D_i[j][i] = integral.imag

    def _integral(self: typing_extensions.Self, i: int, j: int) -> complex:
        """Calculate the D integral between two spherical harmonics."""
        if i not in self._fields:
            self._fields[i] = sleplet.harmonic_methods.invert_flm_boosted(
                sleplet.harmonic_methods._create_spherical_harmonic(self.L, i),
                self.L,
                self._resolution,
            )
        if j not in self._fields:
            self._fields[j] = sleplet.harmonic_methods.invert_flm_boosted(
                sleplet.harmonic_methods._create_spherical_harmonic(self.L, j),
                self.L,
                self._resolution,
            )
        return sleplet._integration_methods.integrate_region_sphere(
            self.mask,
            self._weight,
            self._fields[i],
            self._fields[j].conj(),
        )



"""Contains the `Mesh` class."""

import dataclasses

import numpy as np
import numpy.typing as npt
import plotly.graph_objs as go
import pydantic
import typing_extensions

import sleplet._mesh_methods
import sleplet._plotly_methods
import sleplet._validation


@pydantic.dataclasses.dataclass(config=sleplet._validation.validation)
class Mesh:
    """Create a mesh object."""

    name: str
    """The name of the mesh. Current options `bird`, `cheetah`, `cube`, `dragon`
    `homer` and `teapot`."""
    _: dataclasses.KW_ONLY
    number_basis_functions: int | None = None
    """The number of basis functions to compute from the Laplacian of the given
    mesh. Defaults to one quarter of the number of vertices o the given mesh."""
    zoom: bool = False
    """Whether to zoom in on the pre-selected region of the mesh in the
    plots."""
    _camera_view: go.layout.scene.Camera | None = pydantic.Field(
        default=go.layout.scene.Camera(),
        init_var=False,
        repr=False,
    )
    _colourbar_pos: float = pydantic.Field(default=0, init_var=False, repr=False)
    basis_functions: npt.NDArray[np.float64] = pydantic.Field(
        default_factory=lambda: np.empty(0),
        init_var=False,
        repr=False,
    )
    faces: npt.NDArray[np.float64] = pydantic.Field(
        default_factory=lambda: np.empty(0),
        init_var=False,
        repr=False,
    )
    mesh_eigenvalues: npt.NDArray[np.float64] = pydantic.Field(
        default_factory=lambda: np.empty(0),
        init_var=False,
        repr=False,
    )
    mesh_region: npt.NDArray[np.bool_] = pydantic.Field(
        default_factory=lambda: np.empty(0, dtype=np.bool_),
        init_var=False,
        repr=False,
    )
    vertices: npt.NDArray[np.float64] = pydantic.Field(
        default_factory=lambda: np.empty(0),
        init_var=False,
        repr=False,
    )

    def __post_init__(self: typing_extensions.Self) -> None:
        mesh_config = sleplet._mesh_methods.extract_mesh_config(self.name)
        self._camera_view = sleplet._plotly_methods.create_camera(
            mesh_config["CAMERA_X"],
            mesh_config["CAMERA_Y"],
            mesh_config["CAMERA_Z"],
            mesh_config["REGION_ZOOM"] if self.zoom else mesh_config["DEFAULT_ZOOM"],
            x_center=mesh_config["CENTER_X"] if self.zoom else 0,
            y_center=mesh_config["CENTER_Y"] if self.zoom else 0,
            z_center=mesh_config["CENTER_Z"] if self.zoom else 0,
        )
        self._colourbar_pos = (
            mesh_config["REGION_COLOURBAR_POS"]
            if self.zoom
            else mesh_config["DEFAULT_COLOURBAR_POS"]
        )
        self.vertices, self.faces = sleplet._mesh_methods.read_mesh(mesh_config)
        self.mesh_region = sleplet._mesh_methods.create_mesh_region(
            mesh_config,
            self.vertices,
        )
        (
            self.mesh_eigenvalues,
            self.basis_functions,
            self.number_basis_functions,
        ) = sleplet._mesh_methods.mesh_eigendecomposition(
            self.name,
            self.vertices,
            self.faces,
            number_basis_functions=self.number_basis_functions,
        )



"""Contains the abstract `Fp` class."""

import abc

import numpy as np
import numpy.typing as npt
import pydantic
import typing_extensions

import sleplet._mask_methods
import sleplet._validation
import sleplet.noise
import sleplet.slepian.region
import sleplet.slepian.slepian_functions
import sleplet.slepian.slepian_polar_cap
import sleplet.slepian_methods
from sleplet.functions.coefficients import Coefficients


@pydantic.dataclasses.dataclass(config=sleplet._validation.validation)
class Fp(Coefficients):
    """Abstract parent class to handle Slepian coefficients on the sphere."""

    slepian: sleplet.slepian.slepian_functions.SlepianFunctions | None = pydantic.Field(
        default=None,
        init_var=False,
        repr=False,
    )

    def __post_init__(self: typing_extensions.Self) -> None:
        self.region: sleplet.slepian.region.Region | None = (
            self.region
            if isinstance(self.region, sleplet.slepian.region.Region)
            else sleplet._mask_methods.create_default_region()
        )
        self.slepian = sleplet.slepian_methods.choose_slepian_method(
            self.L,
            self.region,
        )
        super().__post_init__()

    def rotate(  # noqa: D102
        self: typing_extensions.Self,
        alpha: float,  # noqa: ARG002
        beta: float,  # noqa: ARG002
        *,
        gamma: float = 0,  # noqa: ARG002
    ) -> npt.NDArray[np.complex128]:
        msg = "Slepian rotation is not defined"
        raise NotImplementedError(msg)

    def _translation_helper(
        self: typing_extensions.Self,
        alpha: float,
        beta: float,
    ) -> npt.NDArray[np.complex128]:
        return sleplet.slepian_methods._compute_s_p_omega_prime(
            self.L,
            alpha,
            beta,
            self.slepian,
        ).conj()

    def _add_noise_to_signal(
        self: typing_extensions.Self,
    ) -> tuple[npt.NDArray[np.complex128 | np.float64] | None, float | None]:
        """Add Gaussian white noise converted to Slepian space."""
        self.coefficients: npt.NDArray[np.complex128 | np.float64]
        if self.noise is not None:
            unnoised_coefficients = self.coefficients.copy()
            n_p = sleplet.noise._create_slepian_noise(
                self.L,
                self.coefficients,
                self.slepian,
                self.noise,
            )
            snr = sleplet.noise.compute_snr(self.coefficients, n_p, "Slepian")
            self.coefficients = self.coefficients + n_p
            return unnoised_coefficients, snr
        return None, None

    @abc.abstractmethod
    def _create_coefficients(
        self: typing_extensions.Self,
    ) -> npt.NDArray[np.complex128 | np.float64]:
        raise NotImplementedError

    @abc.abstractmethod
    def _create_name(self: typing_extensions.Self) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def _set_reality(self: typing_extensions.Self) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def _set_spin(self: typing_extensions.Self) -> int:
        raise NotImplementedError

    @abc.abstractmethod
    def _setup_args(self: typing_extensions.Self) -> None:
        raise NotImplementedError



"""Contains the `SlepianLimitLatLon` class."""

import numba
import numpy as np
import numpy.linalg as LA  # noqa: N812
import numpy.typing as npt
import platformdirs
import pydantic
import typing_extensions

import pyssht as ssht

import sleplet._array_methods
import sleplet._data.setup_pooch
import sleplet._mask_methods
import sleplet._validation
import sleplet._vars
import sleplet.slepian.region
from sleplet.slepian.slepian_functions import SlepianFunctions


@pydantic.dataclasses.dataclass(config=sleplet._validation.validation, kw_only=True)
class SlepianLimitLatLon(SlepianFunctions):
    """Class to create a limited latitude longitude Slepian region on the sphere."""

    phi_max: float = sleplet._vars.PHI_MAX_DEFAULT
    r"""Maximum \(\phi\) value."""
    phi_min: float = sleplet._vars.PHI_MIN_DEFAULT
    r"""Minimum \(\phi\) value."""
    theta_max: float = sleplet._vars.THETA_MAX_DEFAULT
    r"""Maximum \(\theta\) value."""
    theta_min: float = sleplet._vars.THETA_MIN_DEFAULT
    r"""Minimum \(\theta\) value."""

    def __post_init__(self: typing_extensions.Self) -> None:
        super().__post_init__()

    def _create_fn_name(self: typing_extensions.Self) -> str:
        return f"slepian_{self.region._name_ending}"

    def _create_region(self: typing_extensions.Self) -> "sleplet.slepian.region.Region":
        return sleplet.slepian.region.Region(
            theta_min=self.theta_min,
            theta_max=self.theta_max,
            phi_min=self.phi_min,
            phi_max=self.phi_max,
        )

    def _create_mask(self: typing_extensions.Self) -> npt.NDArray[np.float64]:
        return sleplet._mask_methods.create_mask_region(self.L, self.region)

    def _calculate_area(self: typing_extensions.Self) -> float:
        return (self.phi_max - self.phi_min) * (
            np.cos(self.theta_min) - np.cos(self.theta_max)
        )

    def _create_matrix_location(self: typing_extensions.Self) -> str:
        return (
            f"slepian_eigensolutions_D_{self.region._name_ending}_L{self.L}_N{self.N}"
        )

    def _solve_eigenproblem(
        self: typing_extensions.Self,
    ) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.complex128]]:
        eval_loc = f"{self.matrix_location}_eigenvalues.npy"
        evec_loc = f"{self.matrix_location}_eigenvectors.npy"
        try:
            eigenvalues = np.load(
                sleplet._data.setup_pooch.find_on_pooch_then_local(eval_loc),
            )
            eigenvectors = np.load(
                sleplet._data.setup_pooch.find_on_pooch_then_local(evec_loc),
            )
        except TypeError:
            K = self._create_K_matrix()
            eigenvalues, eigenvectors = self._clean_evals_and_evecs(LA.eigh(K))
            np.save(platformdirs.user_data_path() / eval_loc, eigenvalues)
            np.save(platformdirs.user_data_path() / evec_loc, eigenvectors[: self.N])
        return eigenvalues, eigenvectors

    def _create_K_matrix(  # noqa: N802
        self: typing_extensions.Self,
    ) -> npt.NDArray[np.complex128]:
        """Compute the K matrix."""
        # Compute sub-integral matrix
        G = self._slepian_integral()

        # Compute Slepian matrix
        dl_array = ssht.generate_dl(np.pi / 2, self.L)
        K = self._slepian_matrix(dl_array, self.L, self.L - 1, G)
        sleplet._array_methods.fill_upper_triangle_of_hermitian_matrix(K)

        return K

    def _slepian_integral(self: typing_extensions.Self) -> npt.NDArray[np.complex128]:
        """
        Syntax:
        G = _slepian_integral().

        Output:
        G  =  Sub-integral matrix (obtained after the use of Wigner-D and
        Wigner-d functions in computing the Slepian integral) for all orders

        Description:
        This piece of code computes the sub-integral matrix (obtained after the
        use of Wigner-D and Wigner-d functions in computing the Slepian integral)
        for all orders using the formulation given in "Slepian spatialspectral
        concentration problem on the sphere: Analytical formulation for limited
        colatitude-longitude spatial region" by A. P. Bates, Z. Khalid and R. A.
        Kennedy.
        """
        G = np.zeros((4 * self.L - 3, 4 * self.L - 3), dtype=np.complex128)

        def helper(row: int, col: int, S: float) -> None:
            """Use conjugate symmetry property to reduce the number of iterations."""
            try:
                Q = (1 / (col**2 - 1)) * (
                    np.exp(1j * col * self.theta_min)
                    * (1j * col * np.sin(self.theta_min) - np.cos(self.theta_min))
                    + np.exp(1j * col * self.theta_max)
                    * (np.cos(self.theta_max) - 1j * col * np.sin(self.theta_max))
                )
            except ZeroDivisionError:
                Q = 0.25 * (
                    2 * 1j * col * (self.theta_max - self.theta_min)
                    + np.exp(2 * 1j * col * self.theta_min)
                    - np.exp(2 * 1j * col * self.theta_max)
                )

            G[2 * (self.L - 1) + row, 2 * (self.L - 1) + col] = Q * S
            G[2 * (self.L - 1) - row, 2 * (self.L - 1) - col] = G[
                2 * (self.L - 1) + row,
                2 * (self.L - 1) + col,
            ].conj()

        # for row equal to zero
        S = self.phi_max - self.phi_min
        for col in range(-2 * (self.L - 1), 1):
            helper(0, col, S)

        # for row not equal to zero
        for row in range(-2 * (self.L - 1), 0):
            S = (1j / row) * (
                np.exp(1j * row * self.phi_min) - np.exp(1j * row * self.phi_max)
            )
            for col in range(-2 * (self.L - 1), 2 * (self.L - 1) + 1):
                helper(row, col, S)

        return G

    @staticmethod
    @numba.njit(parallel=True, fastmath=True)
    def _slepian_matrix(
        dl: npt.NDArray[np.float64],
        L: int,
        N: int,
        G: npt.NDArray[np.complex128],
    ) -> npt.NDArray[np.complex128]:
        """
        Syntax:
        K = _slepian_matrix(dl, L, N, G).

        Input:
        G  =  Sub-integral matrix (obtained after the use of Wigner-D and
        Wigner-d functions in computing the Slepian integral) for all orders

        Output:
        K  =  Slepian matrix

        Description:
        This piece of code computes the Slepian matrix using the formulation
        given in "Slepian spatialspectral concentration problem on the sphere:
        Analytical formulation for limited colatitude-longitude spatial region"
        by A. P. Bates, Z. Khalid and R. A. Kennedy.
        """
        K = np.zeros((L**2, L**2), dtype=np.complex128)

        for ell in numba.prange(L):
            for p in range(ell + 1):
                C1 = np.sqrt((2 * ell + 1) * (2 * p + 1)) / (4 * np.pi)

                for m in range(-ell, ell + 1):
                    ind_lm = ell * (ell + 1) + m

                    for q in range(-p, p + 1):
                        ind_pq = p * (p + 1) + q
                        row = m - q
                        C2 = (-1j) ** row
                        ind_r = 2 * N + row

                        for mp in range(-ell, ell + 1):
                            C3 = dl[ell, N + mp, N + m] * dl[ell, N + mp, N]
                            S1 = 0

                            for qp in range(-p, p + 1):
                                col = mp - qp
                                C4 = dl[p, N + qp, N + q] * dl[p, N + qp, N]
                                ind_c = 2 * N + col
                                S1 += C4 * G[ind_r, ind_c]

                            K[ind_lm, ind_pq] += C3 * S1

                        K[ind_lm, ind_pq] *= C1 * C2
        return K

    @staticmethod
    def _clean_evals_and_evecs(
        eigendecomposition: tuple[
            npt.NDArray[np.complex128], npt.NDArray[np.complex128]
        ],
    ) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.complex128]]:
        """Need eigenvalues and eigenvectors to be in a certain format."""
        # access values
        eigenvalues_complex, eigenvectors = eigendecomposition

        # eigenvalues should be real
        eigenvalues = eigenvalues_complex.real

        # Sort eigenvalues and eigenvectors in descending order of eigenvalues
        idx = eigenvalues.argsort()[::-1]
        eigenvalues = eigenvalues[idx]
        eigenvectors = eigenvectors[:, idx].conj().T

        # ensure first element of each eigenvector is positive
        eigenvectors *= np.where(eigenvectors[:, 0] < 0, -1, 1)[:, np.newaxis]

        return eigenvalues, eigenvectors



import typing

import numpy as np
import numpy.typing as npt

if typing.TYPE_CHECKING:
    import sleplet.meshes.mesh

_MACHINE_EPSILON = 1e-14


def clean_evals_and_evecs(
    eigendecomposition: tuple[npt.NDArray[np.complex128], npt.NDArray[np.complex128]],
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.complex128]]:
    """Need eigenvalues and eigenvectors to be in a certain format."""
    # access values
    eigenvalues_complex, eigenvectors = eigendecomposition

    # eigenvalues should be real
    eigenvalues = eigenvalues_complex.real

    # Sort eigenvalues and eigenvectors in descending order of eigenvalues
    idx = eigenvalues.argsort()[::-1]
    eigenvalues = eigenvalues[idx]
    eigenvectors = eigenvectors[:, idx].conj().T

    # ensure first element of each eigenvector is positive
    eigenvectors *= np.where(eigenvectors[:, 0] < 0, -1, 1)[:, np.newaxis]

    # find repeating eigenvalues and ensure orthorgonality
    pairs = np.where(np.abs(np.diff(eigenvalues)) < _MACHINE_EPSILON)[0] + 1
    eigenvectors[pairs] *= 1j

    return eigenvalues, eigenvectors


def compute_mesh_shannon(mesh: "sleplet.meshes.mesh.Mesh") -> int:
    """Compute the effective Shannon number for a region of a mesh."""
    num_basis_fun = mesh.mesh_eigenvalues.shape[0]
    region_vertices = mesh.mesh_region.sum()
    total_vertices = mesh.mesh_region.shape[0]
    return round(region_vertices / total_vertices * num_basis_fun)



import numpy as np
import numpy.typing as npt


def sifting_convolution(
    f_coefficient: npt.NDArray[np.complex128 | np.float64],
    g_coefficient: npt.NDArray[np.complex128 | np.float64],
    *,
    shannon: int | None = None,
) -> npt.NDArray[np.complex128 | np.float64]:
    """Compute the sifting convolution between two multipoles."""
    n = shannon if shannon is not None else np.newaxis
    # change shape if the sizes don't match
    g_reshape = (
        g_coefficient[np.newaxis]
        if len(g_coefficient.shape) < len(f_coefficient.shape)
        else g_coefficient
    )
    return (f_coefficient.T[:n] * g_reshape.conj().T[:n]).T



import typing

import numpy as np
import numpy.typing as npt


def fill_upper_triangle_of_hermitian_matrix(matrix: npt.NDArray[typing.Any]) -> None:
    """Use Hermitian matrix symmetry can avoid repeated calculations."""
    i_upper = np.triu_indices(len(matrix), k=1)
    matrix[i_upper] = matrix.T[i_upper].conj()



"""Contains the `MeshField` class."""

import igl
import numpy as np
import numpy.typing as npt
import pydantic
import typing_extensions

import sleplet._validation
import sleplet.harmonic_methods
from sleplet.meshes.mesh_harmonic_coefficients import MeshHarmonicCoefficients


@pydantic.dataclasses.dataclass(config=sleplet._validation.validation)
class MeshField(MeshHarmonicCoefficients):
    """Create a per-vertex normals field on a given mesh."""

    def __post_init__(self: typing_extensions.Self) -> None:
        super().__post_init__()

    def _create_coefficients(
        self: typing_extensions.Self,
    ) -> npt.NDArray[np.complex128 | np.float64]:
        """Compute field on the vertices of the mesh."""
        field = igl.per_vertex_normals(self.mesh.vertices, self.mesh.faces)[:, 1]
        return sleplet.harmonic_methods.mesh_forward(self.mesh, field)

    def _create_name(self: typing_extensions.Self) -> str:
        return f"{self.mesh.name}_field"

    def _setup_args(self: typing_extensions.Self) -> None:
        if isinstance(self.extra_args, list):
            msg = f"{self.__class__.__name__} does not support extra arguments"
            raise TypeError(msg)



"""Contains the `Gaussian` class."""

import numpy as np
import numpy.typing as npt
import pydantic
import typing_extensions

import pyssht as ssht

import sleplet._string_methods
import sleplet._validation
from sleplet.functions.flm import Flm


@pydantic.dataclasses.dataclass(config=sleplet._validation.validation, kw_only=True)
class Gaussian(Flm):
    r"""Create a Gaussian \(\exp(-\frac{{\ell}^{2}}{2\sigma^{2}})\)."""

    sigma: float = 10
    r"""Sets the \(\sigma\) value."""

    def __post_init__(self: typing_extensions.Self) -> None:
        super().__post_init__()

    def _create_coefficients(
        self: typing_extensions.Self,
    ) -> npt.NDArray[np.complex128 | np.float64]:
        flm = np.zeros(self.L**2, dtype=np.complex128)
        for ell in range(self.L):
            ind = ssht.elm2ind(ell, 0)
            flm[ind] = np.exp(-ell * (ell + 1) / (2 * self.sigma**2))
        return flm

    def _create_name(self: typing_extensions.Self) -> str:
        return (
            f"{sleplet._string_methods._convert_camel_case_to_snake_case(self.__class__.__name__)}"
            f"{sleplet._string_methods.filename_args(self.sigma, 'sig')}"
        )

    def _set_reality(self: typing_extensions.Self) -> bool:
        return True

    def _set_spin(self: typing_extensions.Self) -> int:
        return 0

    def _setup_args(self: typing_extensions.Self) -> None:
        if isinstance(self.extra_args, list):
            num_args = 1
            if len(self.extra_args) != num_args:
                msg = f"The number of extra arguments should be {num_args}"
                raise ValueError(msg)
            self.sigma = np.float_power(10, self.extra_args[0])



"""Classes to create the Slepian regions on the sphere."""

from .region import Region
from .slepian_arbitrary import SlepianArbitrary
from .slepian_limit_lat_lon import SlepianLimitLatLon
from .slepian_polar_cap import SlepianPolarCap

__all__ = [
    "Region",
    "SlepianArbitrary",
    "SlepianLimitLatLon",
    "SlepianPolarCap",
]



import numpy as np

PHI_0 = np.pi
PHI_MAX_DEFAULT = 2 * np.pi
PHI_MIN_DEFAULT = 0.0
RANDOM_SEED = 30
SAMPLING_SCHEME = "MWSS"
SPHERE_UNSEEN = -1.56e30
THETA_0 = 0.0
THETA_MAX_DEFAULT = np.pi
THETA_MIN_DEFAULT = 0



"""Contains the `SphericalHarmonic` class."""

import numpy as np
import numpy.typing as npt
import pydantic
import typing_extensions

import pyssht as ssht

import sleplet._string_methods
import sleplet._validation
import sleplet.harmonic_methods
from sleplet.functions.flm import Flm


@pydantic.dataclasses.dataclass(config=sleplet._validation.validation, kw_only=True)
class SphericalHarmonic(Flm):
    """Create spherical harmonic functions."""

    ell: int = 0
    r"""Degree \(\ell \geq 0\)."""
    m: int = 0
    r"""Order \(\leq |\ell|\)"""

    def _create_coefficients(
        self: typing_extensions.Self,
    ) -> npt.NDArray[np.complex128 | np.float64]:
        ind = ssht.elm2ind(self.ell, self.m)
        return sleplet.harmonic_methods._create_spherical_harmonic(self.L, ind)

    def _create_name(self: typing_extensions.Self) -> str:
        return (
            f"{sleplet._string_methods._convert_camel_case_to_snake_case(self.__class__.__name__)}"
            f"{sleplet._string_methods.filename_args(self.ell, 'l')}"
            f"{sleplet._string_methods.filename_args(self.m, 'm')}"
        )

    def _set_reality(self: typing_extensions.Self) -> bool:
        return False

    def _set_spin(self: typing_extensions.Self) -> int:
        return 0

    def _setup_args(self: typing_extensions.Self) -> None:
        if isinstance(self.extra_args, list):
            num_args = 2
            if len(self.extra_args) != num_args:
                msg = f"The number of extra arguments should be {num_args}"
                raise ValueError(msg)
            self.ell, self.m = self.extra_args

    @pydantic.field_validator("ell")
    def _check_ell(
        cls,  # noqa: ANN101
        v: int,
        info: pydantic.ValidationInfo,
    ) -> int:
        if not isinstance(v, int):
            msg = "ell should be an integer"
            raise TypeError(msg)
        if v < 0:
            msg = "ell should be positive"
            raise ValueError(msg)
        if v >= info.data["L"]:
            msg = "ell should be less than or equal to L"
            raise ValueError(msg)
        return v

    @pydantic.field_validator("m")
    def _check_m(
        cls,  # noqa: ANN101
        v: int,
        info: pydantic.ValidationInfo,
    ) -> int:
        if not isinstance(v, int):
            msg = "m should be an integer"
            raise TypeError(msg)
        if abs(v) > info.data["ell"]:
            msg = "the magnitude of m should be less than ell"
            raise ValueError(msg)
        return v



"""Contains the `DiracDelta` class."""

import numpy as np
import numpy.typing as npt
import pydantic
import typing_extensions

import pyssht as ssht

import sleplet._string_methods
import sleplet._validation
from sleplet.functions.flm import Flm


@pydantic.dataclasses.dataclass(config=sleplet._validation.validation)
class DiracDelta(Flm):
    """Create a Dirac delta."""

    def __post_init__(self: typing_extensions.Self) -> None:
        super().__post_init__()

    def _create_coefficients(
        self: typing_extensions.Self,
    ) -> npt.NDArray[np.complex128 | np.float64]:
        flm = np.zeros(self.L**2, dtype=np.complex128)
        for ell in range(self.L):
            ind = ssht.elm2ind(ell, 0)
            flm[ind] = np.sqrt((2 * ell + 1) / (4 * np.pi))
        return flm

    def _create_name(self: typing_extensions.Self) -> str:
        return sleplet._string_methods._convert_camel_case_to_snake_case(
            self.__class__.__name__,
        )

    def _set_reality(self: typing_extensions.Self) -> bool:
        return True

    def _set_spin(self: typing_extensions.Self) -> int:
        return 0

    def _setup_args(self: typing_extensions.Self) -> None:
        if isinstance(self.extra_args, list):
            msg = f"{self.__class__.__name__} does not support extra arguments"
            raise TypeError(msg)



import functools
import typing

import numpy as np
import numpy.typing as npt

import pyssht as ssht

import sleplet._vars


def calc_integration_weight(L: int) -> npt.NDArray[np.float64]:
    """Compute the spherical Jacobian for the integration."""
    thetas, phis = ssht.sample_positions(
        L,
        Grid=True,
        Method=sleplet._vars.SAMPLING_SCHEME,
    )
    delta_theta = np.ediff1d(thetas[:, 0]).mean()
    delta_phi = np.ediff1d(phis[0]).mean()
    return np.sin(thetas) * delta_theta * delta_phi


def integrate_whole_sphere(
    weight: npt.NDArray[np.float64],
    *functions: npt.NDArray[np.complex128],
) -> complex:
    """Compute the integration for the whole sphere."""
    multiplied_inputs = _multiply_args(*functions)
    return (multiplied_inputs * weight).sum()


def integrate_region_sphere(
    mask: npt.NDArray[np.float64],
    weight: npt.NDArray[np.float64],
    *functions: npt.NDArray[np.complex128 | np.float64],
) -> complex:
    """Compute the integration for a region of the sphere."""
    multiplied_inputs = _multiply_args(*functions)
    return (multiplied_inputs * weight * mask).sum()


def integrate_whole_mesh(
    vertices: npt.NDArray[np.float64],  # noqa: ARG001
    faces: npt.NDArray[np.int_],  # noqa: ARG001
    *functions: npt.NDArray[np.complex128 | np.float64],
) -> float:
    """Compute the integral of functions on the vertices."""
    multiplied_inputs = _multiply_args(*functions)
    return multiplied_inputs.sum()


def integrate_region_mesh(
    mask: npt.NDArray[np.bool_],
    vertices: npt.NDArray[np.float64],  # noqa: ARG001
    faces: npt.NDArray[np.int_],  # noqa: ARG001
    *functions: npt.NDArray[np.complex128 | np.float64],
) -> float:
    """Compute the integral of a region of functions on the vertices."""
    multiplied_inputs = _multiply_args(*functions)
    return (multiplied_inputs * mask).sum()


def _multiply_args(*args: npt.NDArray[typing.Any]) -> npt.NDArray[typing.Any]:
    """Multiply an unknown number of arguments."""
    return functools.reduce((lambda x, y: x * y), args)



import dataclasses
import logging

import numpy as np
import numpy.typing as npt
import pydantic
import typing_extensions

import sleplet._integration_methods
import sleplet._validation
import sleplet.harmonic_methods
from sleplet.meshes.mesh_slepian import MeshSlepian

_logger = logging.getLogger(__name__)


@pydantic.dataclasses.dataclass(config=sleplet._validation.validation)
class MeshSlepianDecomposition:
    mesh_slepian: MeshSlepian
    _: dataclasses.KW_ONLY
    mask: bool = False
    u_i: npt.NDArray[np.complex128 | np.float64] | None = None
    u: npt.NDArray[np.complex128 | np.float64] | None = None
    _method: str = pydantic.Field(default="", init_var=False, repr=False)

    def __post_init__(self: typing_extensions.Self) -> None:
        self._detect_method()

    def decompose(self: typing_extensions.Self, rank: int) -> float:
        """Decompose the signal into its Slepian coefficients via the given method."""
        self._validate_rank(rank)

        match self._method:
            case "harmonic_sum":
                return self._harmonic_sum(rank)
            case "integrate_mesh":
                return self._integrate_mesh(rank)
            case "integrate_region":
                return self._integrate_region(rank)
            case _:
                msg = f"'{self._method}' is not a valid method"
                raise ValueError(msg)

    def decompose_all(
        self: typing_extensions.Self,
        n_coefficients: int,
    ) -> npt.NDArray[np.float64]:
        """Decompose all ranks of the Slepian coefficients."""
        coefficients = np.zeros(n_coefficients)
        for rank in range(n_coefficients):
            coefficients[rank] = self.decompose(rank)
        return coefficients

    def _integrate_region(self: typing_extensions.Self, rank: int) -> float:
        r"""
        F_{p} =
        \frac{1}{\lambda_{p}}
        \int\limits_{R} \dd{x}
        f(x) \overline{S_{p}(x)}.
        """
        s_p = sleplet.harmonic_methods.mesh_inverse(
            self.mesh_slepian.mesh,
            self.mesh_slepian.slepian_functions[rank],
        )
        integration = sleplet._integration_methods.integrate_region_mesh(
            self.mesh_slepian.mesh.mesh_region,
            self.mesh_slepian.mesh.vertices,
            self.mesh_slepian.mesh.faces,
            self.u,
            s_p,
        )
        return integration / self.mesh_slepian.slepian_eigenvalues[rank]

    def _integrate_mesh(self: typing_extensions.Self, rank: int) -> float:
        r"""
        F_{p} =
        \int\limits_{x} \dd{x}
        f(x) \overline{S_{p}(x)}.
        """
        s_p = sleplet.harmonic_methods.mesh_inverse(
            self.mesh_slepian.mesh,
            self.mesh_slepian.slepian_functions[rank],
        )
        return sleplet._integration_methods.integrate_whole_mesh(
            self.mesh_slepian.mesh.vertices,
            self.mesh_slepian.mesh.faces,
            self.u,
            s_p,
        )

    def _harmonic_sum(self: typing_extensions.Self, rank: int) -> float:
        r"""
        F_{p} =
        \sum\limits_{i=0}^{K}
        f_{i} (S_{p})_{i}^{*}.
        """
        return (self.u_i * self.mesh_slepian.slepian_functions[rank]).sum()

    def _detect_method(self: typing_extensions.Self) -> None:
        """Detect what method is used to perform the decomposition."""
        if self.u_i is not None:
            _logger.info("harmonic sum method selected")
            self._method = "harmonic_sum"
        elif self.u is not None:
            if self.mask:
                _logger.info("integrating a region on the mesh method selected")
                self._method = "integrate_region"
            else:
                _logger.info("integrating the whole mesh method selected")
                self._method = "integrate_mesh"
        else:
            msg = (
                "need to pass one off harmonic coefficients, real pixels "
                "or real pixels with a mask"
            )
            raise RuntimeError(msg)

    def _validate_rank(self: typing_extensions.Self, rank: int) -> None:
        """Check the requested rank is valid."""
        if not isinstance(rank, int):
            msg = "rank should be an integer"
            raise TypeError(msg)
        if rank < 0:
            msg = "rank cannot be negative"
            raise ValueError(msg)
        if rank >= self.mesh_slepian.mesh.number_basis_functions:
            msg = (
                "rank should be less than "
                f"{self.mesh_slepian.mesh.number_basis_functions}"
            )
            raise ValueError(msg)



"""Contains the `Earth` class."""

import numpy as np
import numpy.typing as npt
import pydantic
import typing_extensions

import sleplet._data.create_earth_flm
import sleplet._string_methods
import sleplet._validation
from sleplet.functions.flm import Flm


@pydantic.dataclasses.dataclass(config=sleplet._validation.validation)
class Earth(Flm):
    """Create the topographic map of the Earth."""

    def __post_init__(self: typing_extensions.Self) -> None:
        super().__post_init__()

    def _create_coefficients(
        self: typing_extensions.Self,
    ) -> npt.NDArray[np.complex128 | np.float64]:
        return sleplet._data.create_earth_flm.create_flm(
            self.L,
            smoothing=self.smoothing,
        )

    def _create_name(self: typing_extensions.Self) -> str:
        return sleplet._string_methods._convert_camel_case_to_snake_case(
            self.__class__.__name__,
        )

    def _set_reality(self: typing_extensions.Self) -> bool:
        return True

    def _set_spin(self: typing_extensions.Self) -> int:
        return 0

    def _setup_args(self: typing_extensions.Self) -> None:
        if isinstance(self.extra_args, list):
            msg = f"{self.__class__.__name__} does not support extra arguments"
            raise TypeError(msg)



"""Contains the `SquashedGaussian` class."""

import numpy as np
import numpy.typing as npt
import pydantic
import typing_extensions

import sleplet._string_methods
import sleplet._validation
import sleplet._vars
import sleplet.harmonic_methods
from sleplet.functions.flm import Flm


@pydantic.dataclasses.dataclass(config=sleplet._validation.validation, kw_only=True)
class SquashedGaussian(Flm):
    r"""
    Create a squashed Gaussian
    \(\exp(-\frac{1}{2}{(\frac{\theta-\overline{\theta}}{\sigma_{\theta}})}^{2})
    \sin{(\nu_{\phi}\phi)}\).
    """

    freq: float = 0.1
    r"""Sets the \(\nu_{\phi}\) value."""
    t_sigma: float = 1
    r"""Sets the \(\sigma_{\theta}\) value."""

    def __post_init__(self: typing_extensions.Self) -> None:
        super().__post_init__()

    def _create_coefficients(
        self: typing_extensions.Self,
    ) -> npt.NDArray[np.complex128 | np.float64]:
        return sleplet.harmonic_methods._ensure_f_bandlimited(
            self._grid_fun,
            self.L,
            reality=self.reality,
            spin=self.spin,
        )

    def _create_name(self: typing_extensions.Self) -> str:
        return (
            f"{sleplet._string_methods._convert_camel_case_to_snake_case(self.__class__.__name__)}"
            f"{sleplet._string_methods.filename_args(self.t_sigma, 'tsig')}"
            f"{sleplet._string_methods.filename_args(self.freq, 'freq')}"
        )

    def _set_reality(self: typing_extensions.Self) -> bool:
        return True

    def _set_spin(self: typing_extensions.Self) -> int:
        return 0

    def _setup_args(self: typing_extensions.Self) -> None:
        if isinstance(self.extra_args, list):
            num_args = 2
            if len(self.extra_args) != num_args:
                msg = f"The number of extra arguments should be {num_args}"
                raise ValueError(msg)
            self.t_sigma, self.freq = (np.float_power(10, x) for x in self.extra_args)

    def _grid_fun(
        self: typing_extensions.Self,
        theta: npt.NDArray[np.float64],
        phi: npt.NDArray[np.float64],
    ) -> npt.NDArray[np.float64]:
        """Define the function on the grid."""
        return np.exp(
            -(((theta - sleplet._vars.THETA_0) / self.t_sigma) ** 2) / 2,
        ) * np.sin(self.freq * phi)
